/*
 *  Program in cgi-bin to calculate the mandelbrot fractal
 * 
 *  Copyright (c) 1995 Fabrice Bellard
 *
 *  Contact addresses:
 *  mail: Fabrice Bellard, 451 chemin du mas de Matour, 34790 Grabels, France
 *  email: bellard@poly.polytechnique.fr
 *  url: http://www.polytechnique.fr/poly/~bellard
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <string.h>
#include <fcntl.h>


#include <stdlib.h>
#include <stdio.h>


#define MANDEL_COUNT "/users/eleve/www/lib/Count/Users/mandel_zoom"
#define MANDEL_ERROR "/users/cie6/bellard/Mandelbrot/mandel_error.gif"
#define ZOOMFACTOR 3.0
#define MANDEL_PORT "6620"
#define MANDEL_HOME "/poly/~bellard/mandelbrot/"
#define MANDEL_CGI  "/poly1/cgi-bin/mandel_zoom"

#define BUFSIZE 4096
#define HOST_NB 4

char *mandel_host[]= {
	 "rotule",
	 "tarse",
	 "femur",
	 "tibia",
};



int read_count(void) {
  FILE *f;
  int count;
  
  f=fopen(MANDEL_COUNT,"r");
  if (f==NULL) return 0;
  fscanf(f,"%d",&count);
  fclose(f);
  return count;
}

void write_count(int count) {
  FILE *f;
  f=fopen(MANDEL_COUNT,"w");
  fprintf(f,"%d",count);
  fclose(f);
}

int connect_socket(char *host,char *port) 
{
  int s,service_num;
  char addr[4];
  struct hostent *serverhost;
  struct hostent host_hack;
  struct sockaddr_in serveraddr;
  struct servent *service;
  char *addr_hack;
  
  s = socket (AF_INET, SOCK_STREAM, 0);
  if (s == -1) return -1;
  if (host[0] >= '0' && host[0] <= '9'){
    int a0, a1, a2, a3;
	   sscanf (host, "%d.%d.%d.%d", &a0, &a1, &a2, &a3);
	   addr [0] = a0; addr [1] = a1; addr [2] = a2; addr [3] = a3;
	   serverhost = gethostbyaddr (addr, 4, AF_INET);
	   if (serverhost == NULL){
	     serverhost = &host_hack;
	     serverhost->h_addrtype = AF_INET;
	     addr_hack = addr;
	     serverhost->h_addr_list = &addr_hack;
	     serverhost->h_length = 4;
	   }
	 } else {
	   serverhost = gethostbyname (host);
	   if (serverhost == NULL) return -1;
	 }
	 
	 if (port[0] >= '0' && port[0] <= '9'){
			service_num = htons (atoi (port));
	 } else {
			service = getservbyname (port, NULL);
			if (service == NULL) return -1;
			service_num = service->s_port;
	 }

	 serveraddr.sin_family = serverhost->h_addrtype;
	 serveraddr.sin_port = service_num;
	 bcopy (serverhost->h_addr_list [0], &serveraddr.sin_addr,
					serverhost->h_length);
	 if (connect (s, (struct sockaddr *) &serveraddr, sizeof (serveraddr)) == -1) return -1;

	 return s;
}


int main(void) {
	 
	 char *query,*path;
	 double xmin,xmax,ymin,ymax,xs,ys,xc,yc;
	 int xsize,ysize,xpos,ypos,count;
	 int s,len;
	 char buf[BUFSIZE];
	 FILE *f;
	 int host_n;
	 
	 query=getenv("QUERY_STRING");
	 path=getenv("PATH_INFO");
	 
	 if (strlen(path)==0) {
			xmin=-2.5;
			ymin=-1.5;
			xmax=1.5;
			ymax=1.5;
			xsize=160;
			ysize=120;
	 } else {
			sscanf(path+5,"%d,%d,%lf,%lf,%lf,%lf",
						 &xsize,&ysize,&xmin,&ymin,&xmax,&ymax);
	 }
	 
	 if (strlen(query)==0) {
	   /* mise  jour compteur */
	   count=read_count();
	   count++;
	   write_count(count);
	   
	   /* envoi de l'image gif */
	   printf("Content-type: image/gif\n\n");

	   if ( xsize<10 || xsize>1024 ||
		ysize<10 || ysize>1024 ||
		xmin<-10 || xmax>10 || xmin>=xmax ||
		ymin<-10 || ymax>10 || ymin>=ymax ) goto gif_error;
	     
	   host_n=0;
	   while (host_n<HOST_NB) {
	     s=connect_socket(mandel_host[host_n],MANDEL_PORT);
	     if (s!=-1) break;
	     host_n++;
	   }
	   if ( s==-1 ) {
	   gif_error:

	     f=fopen(MANDEL_ERROR,"r");

	     while (1) {
	       len=fread(buf,1,BUFSIZE,f);
	       if (len==0) break;
	       fwrite(buf,1,len,stdout);
	     } 

	     fclose(f);
	   } else {

	     sprintf(buf,"%d %d %0.10f %0.10f %0.10f %0.10f\n",
		     xsize,ysize,xmin,ymin,xmax,ymax);
	     write(s,buf,strlen(buf));
	     
	     while (1) {
	       len=read(s,buf,BUFSIZE);
	       if (len==0) break;
	       fwrite(buf,1,len,stdout);
	     } 
	     
	     close(s);
	   }
	 } else {
	   /* lecture de count */
	   count=read_count();

	   /* envoi de la page html */
			sscanf(query,"%d,%d",&xpos,&ypos);
			if (xpos>=0 && xpos<xsize && ypos>=0 && ypos<ysize)  {
				 
				 xc=xmin+(xmax-xmin)*(double)(xpos)/(double)(xsize);
				 yc=ymin+(ymax-ymin)*(double)(ypos)/(double)(ysize);
				 
				 xs=(xmax-xmin)/ZOOMFACTOR/2.0;
				 ys=(ymax-ymin)/ZOOMFACTOR/2.0;
			
				 xmin=xc-xs;
				 xmax=xc+xs;
				 ymin=yc-ys;
				 ymax=yc+ys;
			}
			
			printf("Content-type: text/html\n\n");
			
			printf("<HTML><TITLE>Zoom on the Mandelbrot Set</TITLE>"
			       "<CENTER><H1><B>Zoom on the Mandelbrot Set</B></H1></CENTER><P>");

/*
			printf("query=%s path=%s xpos=%d ypos=%d <P>",query,path,xpos,ypos);
*/

						 
			printf("<P><B>Please click on the zone you want to zoom into.</B><P>");
			
			printf("<A HREF=""%s/gif,%d,%d,%0.10f,%0.10f,%0.10f,%0.10f"">",
						 MANDEL_CGI,xsize,ysize,xmin,ymin,xmax,ymax);
			
			printf("<CENTER><IMG ALIGN=middle SRC=""%s/gif,%d,%d,%0.10f,%0.10f,%0.10f,%0.10f"" ISMAP> </CENTER></A>\n",
			       MANDEL_CGI,xsize,ysize,xmin,ymin,xmax,ymax);

printf("<HR>Image size= %dx%d pixels. Zoom factor= %0.1f. Zoom #%d.",
       xsize,ysize,ZOOMFACTOR,count);

printf(
"<BR><A HREF=""%s""> Return to the Mandelbrot home page </A>"
"<HR><H5>by Fabrice Bellard (<A HREF=""mailto:bellard@poly.polytechnique.fr"">"
"bellard@poly.polytechnique.fr</A>)</H5>"
"</HTML>",MANDEL_HOME);
	 }
	 
	 return 0;
}
	 
